#!/usr/bin/env python3
from time import time
from shutil import rmtree
from traceback import format_exc

from .database import Edge
from .database import Vertex
from .database import Database


tests = list()


def test(func):
    tests.append(func)


@test
def create():
    graph = Database('/tmp/ajgu')
    graph.close()
    return True


@test
def create_vertex():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        v = txn.vertex('new')
    return v


@test
def create_and_get_vertex():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        v = txn.vertex('get-vertex')
    with graph.transaction() as txn:
        v = txn.get(v.identifier())
    return v


@test
def create_modify_and_get_vertex():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        v = txn.vertex('create-modify-and-get-vertex')
        v['value'] = 'key'
    with graph.transaction() as txn:
        v = txn.get(v.identifier())
    r = v['value'] == 'key'
    return r


@test
def create_and_get_modify_and_get_again_vertex():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        v = txn.vertex('create-and-get-modify-and-get-again-vertex')
    with graph.transaction() as txn:
        v = txn.get(v.identifier())
        v['value'] = 'key'
    with graph.transaction() as txn:
        v = txn.get(v.identifier())
    r = v['value'] == 'key'
    return r


@test
def create_modify_and_get_edge():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        start = txn.vertex('start')
        end = txn.vertex('end')
        edge = txn.edge(start, 'héllo', end)
        edge['who'] = 'world'
    with graph.transaction() as txn:
        edge = txn.get(edge.identifier())
    r = edge['who'] == 'world'
    return r


@test
def create_edge_and_checkout_vertex():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        start = txn.vertex('start')
        end = txn.vertex('end')
        edge = txn.edge(start, 'héllo', end)
    # retrieve start and end
    with graph.transaction() as txn:
        start = txn.get(start.identifier())
        end = txn.get(end.identifier())
        r = list(start.outgoings()) and list(end.incomings())
    return r


@test
def create_edge_and_filtered_edges():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        # vertices
        start = txn.vertex('start')
        end = txn.vertex('end')
        # first edge
        edge = txn.edge(start, 'héllo', end)
        # second edge
        edge = txn.edge(start, 'world', end)
    # retrieve start and end
    with graph.transaction() as txn:
        start = txn.get(start.identifier())
        end = txn.get(end.identifier())
    assert len(list(start.outgoings())) == 2
    assert len(list(end.incomings('héllo'))) == 1
    assert len(list(end.incomings('world'))) == 1
    assert len(list(start.outgoings('héllo'))) == 1
    assert len(list(start.outgoings('world'))) == 1
    return True



@test
def delete_node():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        v = txn.vertex('delete-node')
        v['value'] = 'key'
    # retrieve it
    with graph.transaction() as txn:
        v = txn.vertex(v.identifier())
    if not v:
        return False
    # delete it
    with graph.transaction() as txn:
        v.delete()
    with graph.transaction() as txn:
        try:
            v = txn.get(v.identifier())
        except:
            return True
        else:
            return False


@test
def delete_edge():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        # vertices
        start = txn.vertex('start')
        end = txn.vertex('end')
        # first edge
        edge = txn.edge(start, 'héllo', end)

    # delete edge
    with graph.transaction() as txn:
        edge = txn.get(edge.identifier())
        edge.delete()

    # check end and start node were updated
    with graph.transaction() as txn:
        start = txn.get(start.identifier())
        if len(list(start.outgoings())) != 0:
            return False
        end = txn.get(end.identifier())
        if len(list(end.incomings())) != 0:
            return False
    return True


@test
def iterate_over_all_elements():
    graph = Database('/tmp/ajgu')
    with graph.transaction() as txn:
        one = txn.vertex('one')
        two = txn.vertex('two')
        three = txn.vertex('three')
        edge = txn.edge(one, 'one->three', three)
    expected = dict()
    expected[one.identifier()] = Vertex
    expected[two.identifier()] = Vertex
    expected[three.identifier()] = Vertex
    expected[edge.identifier()] = Edge
    with graph.transaction() as txn:
        elements = txn.all()
    for index, element in enumerate(elements):
        try:
            ElementClass = expected[element.identifier()]
        except KeyError:
            msg = 'Element with id=%s not found (%s)'
            msg = msg % (element.identifier(), index)
            print(msg)
            return False
        else:
            if not isinstance(element, ElementClass):
                msg = 'Element with id=%s MUST be %s (%s)'
                msg = msg % (element.identifier(), ElementClass, index)
                print(msg)
                return False
    return True


# @test
# def filter_query():
#     database = Database('/tmp/ajgu')
#     with database.transaction() as txn:
#         user = txn.vertex('user')
#         user['name'] = 'amirouche'
#         for i in range(10):
#             url = txn.vertex('url')
#             url['url'] = 'http://hypermove.net'
#             txn.edge(user, 'likes', url)
#         # spare outgoings edges from amirouche
#         for i in range(10):
#             vertex = txn.vertex(str(i))
#             label = 'amirouche->%s' % str(i)
#             txn.edge(user, label, vertex)
#         # spare incomings edges to amirouche
#         for i in range(10):
#             label = '%s->amirouche' % str(i)
#             vertex = txn.vertex(str(i))
#             txn.edge(vertex, label, user)
#         # spare urls
#         for i in range(10):
#             txn.vertex('url')
#         # spare users
#         for i in range(10):
#             txn.vertex('user')
#     with database.transaction() as txn:
#         filter_specification = ('name', 'eq', 'amirouche')
#         query = txn.query('testing amirouche').label('user')
#         query = query.filter(filter_specification)
#         user = query.one()
#         assert user
#         urls = query.outgoings.label('likes').end.all()
#         for url in urls:
#             assert url.get('url', None) == 'http://hypermove.net'
#     return True


#@test  # XXX: example usage
def create_feeds():
    database = Database('/tmp/ajgu')
    # create user
    with database.transaction() as txn:
        amirouche = txn.vertex('user')
        amirouche['name'] = 'amirouche'

    # create feed
    with database.transaction() as txn:
        amirouche = txn.get(amirouche.identifier())

        hypermove = txn.vertex('feed')
        hypermove['url'] = 'http://hyperdev.fr/atom.xml'
        txn.edge(amirouche, 'feeds', hypermove)

        hypermove = txn.vertex('feed')
        hypermove['url'] = 'http://www.hyperdev.fr/keyword/fran%C3%A7ais/atom.xml'
        txn.edge(amirouche, 'feeds', hypermove)

    with database.transaction() as txn:
        amirouche = txn.get(amirouche.identifier())
        for feed in map(lambda x: x.end(), amirouche.outgoings('feeds')):
            from requests import get
            from lxml.etree import fromstring
            xml = fromstring(get(feed.url).text)
            ns = dict(a='http://www.w3.org/2005/Atom')
            list(map(
                print,
                xml.xpath('/a:feed/a:entry/a:title/text()', namespaces=ns)
            ))
    return True


@test
def run_uzelmumu_query():
    # populate database
    database = Database('/tmp/ajgu')
    # create user
    with database.transaction() as txn:
        amirouche = txn.vertex('user')
        amirouche['name'] = 'amirouche'

    # create feed
    with database.transaction() as txn:
        amirouche = txn.get(amirouche.identifier())

        hypermove = txn.vertex('feed')
        hypermove['url'] = 'http://hyperdev.fr/atom.xml'
        txn.edge(amirouche, 'feeds', hypermove)

        hypermove = txn.vertex('feed')
        url = 'http://www.hyperdev.fr/keyword/fran%C3%A7ais/atom.xml'
        hypermove['url'] = url
        txn.edge(amirouche, 'feeds', hypermove)

    from uzelmumu import graph
    from uzelmumu import run

    query = graph().get(amirouche.identifier()).outgoings('feeds')
    query = query.end().values()
    query = query.serialize()

    from database import dumps
    from database import loads

    query = loads(dumps(query))

    feeds = run(database, query)

    urls = list()
    for feed in feeds:
        urls.append(feed['properties']['url'])
    if not (urls[0] == 'http://hyperdev.fr/atom.xml'):
        return False
    if not (urls[1] == 'http://www.hyperdev.fr/keyword/fran%C3%A7ais/atom.xml'):
        return False
    return True

# @test
def order_query_without_index():
    database = Database('/tmp/ajgu')
    with database.transaction() as txn:
        user = txn.vertex('user')
        user['name'] = 'amirouche'
        for i in range(10):
            vertex = txn.vertex()
            vertex['index'] = i
            label = 'relevant'
            edge = txn.edge(user, label, vertex)
            edge['index'] = i
        # spare edges
        for i in range(10):
            vertex = txn.vertex()
            label = 'irrelevant'
            edge = txn.edge(user, label, vertex)
    with database.transaction() as txn:
        filter_specification = ('name', 'eq', 'amirouche')
        query = txn.query('retrieve ordered edges').label('user')
        query = query.filter(filter_specification)
        edges = query.outgoing('relevant').order_by('index').all()
        indexes = list(map(lambda x: x['index'], edges))
        expected = sorted(indexes)
        assert indexes == expected
    return True


# @test
# def index_vs_no_label_index_timing():
#     database = Database('/tmp/ajgu')
#     with database.transaction() as txn:
#         for i in range(1000):
#             txn.vertex('not-indexed')
#     with database.transaction() as txn:
#         txn.index('indexed').label('indexed')
#         for i in range(1000):
#             txn.vertex('indexed')
#     with database.transaction() as txn:
#         # indexed timing
#         start = time()
#         txn.query.label('indexed').all()
#         indexed = time() - start
#         # not indexed timing
#         start = time()
#         txn.query.label('not-indexed').all()
#         not_indexed = time() - start
#     assert index > not_indexed
#     return True


if __name__ == '__main__':
    for func in tests:
        try:
            r = func()
        except Exception as e:
            print('* %s: ERROR :(' % func.__name__)
            print(format_exc())
        else:
            if r:
                print('* %s: PASSED :)' % func.__name__)
            else:
                print('* %s: FAILED :(' % func.__name__)
        finally:
            rmtree('/tmp/ajgu')
